package com.massa.mrules.demos.sudoku.java;

import com.massa.log.Log;
import com.massa.log.LogFactory;
import com.massa.mrules.addon.GlobalCacheActivationResolver;
import com.massa.mrules.context.compile.CompilationLevel;
import com.massa.mrules.context.execute.MExecutionContext;
import com.massa.mrules.exception.MConfigurationException;
import com.massa.mrules.exception.MExecutionException;
import com.massa.mrules.executable.IExecutionResult;
import com.massa.mrules.extensions.dsl.factory.NativeLanguageFactory;
import com.massa.mrules.factory.IRuleExecutionSetFactory;
import com.massa.mrules.factory.ser.SerializationFactory;
import com.massa.mrules.factory.xml.RichXMLFactory;
import com.massa.mrules.factory.xml.XMLMConfigurationException;
import com.massa.mrules.set.IMRuleExecutionSet;
import com.massa.mrules.set.MRuleExecutionSet;
import com.massa.util.LicenceInfoHelper;
import com.massa.util.UtilsException;
import com.massa.util.classloader.ResourceLoader;
import com.massa.util.convert.ConvertUtils;

import org.junit.Assert;
import org.junit.Assume;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Arrays;
import java.util.Collection;
import java.util.Map;

/**
 * Test for MRules solver.
 */
@RunWith(Parameterized.class)
public class MRulesTest extends AbstractSudoku {

	private static final Log LOG = LogFactory.getLog(MRulesTest.class);

	private static final String[] CONFIG = {"mrules.sudoku.xml", "mrules.sudoku2.xml", "mrules.sudoku.mrl"};

	private final int version;
	private final CompilationLevel level;
	private final boolean activateOptimization;
	private final boolean cacheEnabled;

	private final IMRuleExecutionSet ruleset;

	@Parameters(name="V{0} - level: {1} - optim: {2} - cache: {3}")
	public static Collection<Object[]> configs() {
		return Arrays.asList(new Object[][] {
			{1, CompilationLevel.STANDARD, true, true},
			{1, CompilationLevel.LIGHT, true, true},
			{1, CompilationLevel.STANDARD, false, true},
			{1, CompilationLevel.LIGHT, false, true},
			{1, CompilationLevel.STANDARD, true, false},
			{1, CompilationLevel.LIGHT, true, false},
			{1, CompilationLevel.STANDARD, false, false},
			{1, CompilationLevel.LIGHT, false, false},

			{2, CompilationLevel.STANDARD, true, true},
			{2, CompilationLevel.LIGHT, true, true},
			{2, CompilationLevel.STANDARD, false, true},
			{2, CompilationLevel.LIGHT, false, true},
			{2, CompilationLevel.STANDARD, true, false},
			{2, CompilationLevel.LIGHT, true, false},
			{2, CompilationLevel.STANDARD, false, false},
			{2, CompilationLevel.LIGHT, false, false},

			{3, CompilationLevel.STANDARD, true, true},
			{3, CompilationLevel.LIGHT, true, true},
			{3, CompilationLevel.STANDARD, false, true},
			{3, CompilationLevel.LIGHT, false, true},
			{3, CompilationLevel.STANDARD, true, false},
			{3, CompilationLevel.LIGHT, true, false},
			{3, CompilationLevel.STANDARD, false, false},
			{3, CompilationLevel.LIGHT, false, false},
		});
	}

	public MRulesTest(final int version, final CompilationLevel level, final boolean activateOptimization, final boolean cacheEnabled)
			throws Exception {
		Assume.assumeFalse("DSL is not licenced.", MRulesTest.CONFIG[version-1].endsWith("mrl") && !LicenceInfoHelper.isModuleLicenced("DSL"));

		this.version = version - 1;
		this.level = level;
		this.activateOptimization = activateOptimization;
		this.cacheEnabled = cacheEnabled;

		this.ruleset = this.getEng(MRulesTest.CONFIG[this.version]);
	}

	@Test
	public void testDroolsExamples() throws MExecutionException, UtilsException {
		final MExecutionContext<int[][]> ctxt = new MExecutionContext<>(this.ruleset);
		for (final Map.Entry<String, int[][]> model: this.getDroolsExamples().entrySet()) {
			this.dump(model.getKey(), model.getValue());

			ctxt.setInput(model.getValue());
			final IExecutionResult result = ctxt.execute();
			final boolean solved = ConvertUtils.getDefaultInstance().cast(boolean.class, result.getReturnedValue());

			this.dump("By mrules: " + model.getKey(), model.getValue());

			Assert.assertTrue(solved ^ "!DELIBERATELY BROKEN!".equals(model.getKey()));
		}
	}

	@Test
	public void testSimple() throws MExecutionException, UtilsException {
		this.ptest(this.getSimpleModel());
	}

	@Test
	public void testMepham() throws MExecutionException, UtilsException {
		this.ptest(this.getMephamDiabolicalModel());
	}

	private void ptest(final int[][] model) throws MExecutionException, UtilsException {

		final MExecutionContext<int[][]> ctxt = new MExecutionContext<>(this.ruleset, model);
		final IExecutionResult result = ctxt.execute();
		this.dump("origin", model);
		final boolean solved = ConvertUtils.getDefaultInstance().cast(boolean.class, result.getReturnedValue());
		Assert.assertTrue(solved);
	}

	@Test
	public void testSimplePerf() throws MExecutionException {
		final int[][][] models = new int[20][][];
		for (int i = 0 ; i < models.length ; i++) {
			models[i] = this.getSimpleModel();
		}
		this.ptestPerf(models);
	}

	@Test
	public void testMephamPerf() throws MExecutionException {
		final int[][][] models = new int[4][][];
		for (int i = 0 ; i < models.length ; i++) {
			models[i] = this.getMephamDiabolicalModel();
		}
		this.ptestPerf(models);
	}

	public void ptestPerf(final int[][][] models) throws MExecutionException {
		final MExecutionContext<int[][]> ctxt = new MExecutionContext<>(this.ruleset);
		//to avoid debug and info.
		ctxt.setLoggerName(".perf");
		ctxt.setOut(null);
		final int warmupEnd = models.length/2;

		MRulesTest.LOG.info("Warmup ...");
		for (int i = 0 ; i < warmupEnd ; i++) {
			ctxt.setInput(models[i]);
			ctxt.execute();
		}
		MRulesTest.LOG.info("Starting MRules ...");
		final long t0 = System.currentTimeMillis();
		for (int i = warmupEnd ; i < models.length ; i++) {
			ctxt.setInput(models[i]);
			ctxt.execute();
		}
		MRulesTest.LOG.info("Ending MRules in : " + (System.currentTimeMillis() - t0) + " ms.");
	}

	@Test
	public void testReadWrite() throws Exception {
		this.pTestReadWrite(MRulesTest.CONFIG[this.version]);
	}

	private void pTestReadWrite(final String name) throws Exception {
		final StringWriter w = new StringWriter();
		final RichXMLFactory factory = new RichXMLFactory(this.level);
		factory.write(w, this.ruleset);
		MRulesTest.LOG.info(w.toString());

		final ByteArrayOutputStream o = new ByteArrayOutputStream();
		final SerializationFactory sfactory = new SerializationFactory(this.level);
		sfactory.write(o, this.ruleset);

		Assert.assertEquals(this.ruleset, this.ruleset.clone());
		final IMRuleExecutionSet readFromXml = factory.read(new StringReader(w.toString()));
		Assert.assertEquals(this.ruleset.toString(), readFromXml.toString());
		Assert.assertTrue(((MRuleExecutionSet) this.ruleset).equals(readFromXml));
		Assert.assertEquals(this.ruleset, sfactory.read(new ByteArrayInputStream(o.toByteArray())));
	}


	protected IMRuleExecutionSet getEng(final String name) throws MConfigurationException, XMLMConfigurationException {
		final InputStream config = ResourceLoader.getResourceAsStream(name);
		final long t0 = System.currentTimeMillis();
		final IRuleExecutionSetFactory factory = this.getFactory(name);
		final long t1 = System.currentTimeMillis();
		MRulesTest.LOG.info("Getting factory in " + (t1-t0) + " ms.");
		final IMRuleExecutionSet eng = factory.read(config);
		final long t2 = System.currentTimeMillis();
		MRulesTest.LOG.info("Getting ruleset in " + (t2-t1) + " ms.");

		return eng;
	}

	private IRuleExecutionSetFactory getFactory(final String name) throws MConfigurationException {
		final IRuleExecutionSetFactory factory;
		if (name.endsWith("xml")) {
			factory = new RichXMLFactory(this.level);
		} else if (name.endsWith("mrl")) {
			factory = new NativeLanguageFactory(this.level);
		} else {
			throw new RuntimeException("Unknown extension");
		}
		if (!this.cacheEnabled) {
			factory.setCacheActivationResolver(new GlobalCacheActivationResolver(false));
		}
		factory.setOptimizationEnabled(this.activateOptimization);
		return factory;
	}
}
