/*
 * Copyright 2011 Red Hat, Inc. and/or its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.drools.examples.sudoku;

import java.util.Formatter;
import java.util.HashSet;
import java.util.Set;

import org.kie.api.runtime.KieContainer;
import org.kie.api.runtime.KieSession;
import org.kie.api.runtime.rule.FactHandle;

/**
 * An object of this class solves Sudoku problems.
 */
public class Sudoku {

	public  Cell[][]    cells;
	private final CellSqr[][] sqrs = new CellSqr[][]{ new CellSqr[3], new CellSqr[3], new CellSqr[3] };
	private final CellRow[]   rows = new CellRow[9];
	private final CellCol[]   cols = new CellCol[9];

	private final KieContainer kc;
	private KieSession session;
	private Counter counter;
	private Boolean explain = false;
	private FactHandle steppingFactHandle;
	private Stepping stepping;
	private boolean unsolvable = false;

	/**
	 * Constructor.
	 */
	public Sudoku(final KieContainer kc) {
		this.kc = kc;
	}

	public String getCellValue(final int iRow, final int iCol) {
		if (this.cells == null) {
			return " ";
		}
		return this.cells[iRow][iCol].valueAsString();
	}

	/**
	 * Nice printout of the grid.
	 */
	public void dumpGrid() {
		final Formatter fmt = new Formatter(System.out);
		fmt.format("       ");
		for (int iCol = 0; iCol < 9; iCol++) {
			fmt.format("Col: %d     ", iCol);
		}
		System.out.println();
		for (int iRow = 0; iRow < 9; iRow++) {
			System.out.print("Row " + iRow + ": ");
			for (int iCol = 0; iCol < 9; iCol++) {
				if (this.cells[iRow][iCol].getValue() != null) {
					fmt.format(" --- %d --- ", this.cells[iRow][iCol].getValue());
				} else {
					final StringBuilder sb = new StringBuilder();
					final Set<Integer> perms = this.cells[iRow][iCol].getFree();
					for (int i = 1; i <= 9; i++) {
						if (perms.contains(i)) {
							sb.append(i);
						} else {
							sb.append(' ');
						}
					}
					fmt.format(" %-10s", sb.toString());
				}
			}
			System.out.println();
		}
		fmt.close();
	}

	/**
	 * Checks that everything is still according to the sudoku rules.
	 */
	public void consistencyCheck() {
		for (int iRow = 0; iRow < 9; iRow++) {
			for (int iCol = 0; iCol < 9; iCol++) {
				final Cell cell = this.cells[iRow][iCol];
				final Integer value = cell.getValue();
				if (value != null) {
					if (! cell.getFree().isEmpty()) {
						throw new IllegalStateException("free not empty");
					}
					// any containing group
					for (final Cell other: cell.getExCells()) {
						// must not occur in any of the other cells
						if (value.equals(other.getValue())) {
							throw new IllegalStateException("duplicate");
						}
						// must not occur in the permissibles of any of the other cells
						if (other.getFree().contains(value)) {
							throw new IllegalStateException("not eliminated");
						}
					}
				}
			}
		}

		for (int i = 0; i < this.rows.length; i++) {
			final Set<Integer> aSet = new HashSet<Integer>();
			for (int j = 0; j < this.rows[i].getCells().size(); j++) {
				final Cell cell = this.rows[i].getCells().get(j);
				final Integer value = cell.getValue();
				if (value != null) {
					aSet.add(value);
				} else {
					aSet.addAll(cell.getFree());
				}
			}
			if (! aSet.equals(CellGroup.ALL_NINE)) {
				throw new IllegalStateException("deficit in row");
			}
		}

		for (int i = 0; i < this.cols.length; i++) {
			final Set<Integer> aSet = new HashSet<Integer>();
			for (int j = 0; j < this.cols[i].getCells().size(); j++) {
				final Cell cell = this.cols[i].getCells().get(j);
				final Integer value = cell.getValue();
				if (value != null) {
					aSet.add(value);
				} else {
					aSet.addAll(cell.getFree());
				}
			}
			if (! aSet.equals(CellGroup.ALL_NINE)) {
				throw new IllegalStateException("deficit in column");
			}
		}

		for (int ir = 0; ir < this.sqrs.length; ir++) {
			for (int ic = 0; ic < this.sqrs[ir] .length; ic++) {
				final Set<Integer> aSet = new HashSet<Integer>();
				for (int j = 0; j < this.sqrs[ir][ic].getCells().size(); j++) {
					final Cell cell = this.sqrs[ir][ic].getCells().get(j);
					final Integer value = cell.getValue();
					if (value != null) {
						aSet.add(value);
					} else {
						aSet.addAll(cell.getFree());
					}
				}
				if (! aSet.equals(CellGroup.ALL_NINE)) {
					throw new IllegalStateException("deficit in square");
				}
			}
		}
		System.out.println("+++ check OK +++");
	}

	public void solve() {
		if (this.isSolved()) {
			return;
		}
		this.explain = false;
		this.session.setGlobal("explain", this.explain);
		if( this.steppingFactHandle != null ){
			this.session.delete( this.steppingFactHandle );
			this.steppingFactHandle = null;
			this.stepping = null;
		}
		this.session.fireAllRules();
	}

	public void step() {
		if (this.isSolved()) {
			return;
		}
		this.explain = true;
		this.session.setGlobal("explain", this.explain);
		this.counter.setCount(1);
		this.session.update(this.session.getFactHandle(this.counter), this.counter);
		if( this.steppingFactHandle == null ){
			this.steppingFactHandle = this.session.insert( this.stepping = new Stepping() );
		}
		this.session.fireUntilHalt();
		if( this.stepping.isEmergency() ){
			this.unsolvable = true;
		}
	}

	public boolean isSolved() {
		for (int iRow = 0; iRow < 9; iRow++) {
			for (int iCol = 0; iCol < 9; iCol++) {
				if (this.cells[iRow][iCol].getValue() == null) {
					return false;
				}
			}
		}
		return true;
	}

	public boolean isUnsolvable(){
		return this.unsolvable;
	}

	private void create() {
		for (int i = 0; i < 9; i++) {
			this.session.insert(Integer.valueOf(i+1));
			this.rows[i] = new CellRow(i);
			this.cols[i] = new CellCol(i);
		}

		this.cells = new Cell[9][];
		for (int iRow = 0; iRow < 9; iRow++) {
			this.cells[iRow] = new Cell[9];
			for (int iCol = 0; iCol < 9; iCol++) {
				final Cell cell = this.cells[iRow][iCol] = new Cell();
				this.rows[iRow].addCell(cell);
				this.cols[iCol].addCell(cell);
			}
		}

		for (int i = 0; i < 3; i++) {
			for (int j = 0; j < 3; j++) {
				this.sqrs[i][j] = new CellSqr(this.rows[i*3], this.rows[i*3+1], this.rows[i*3+2],
						this.cols[j*3], this.cols[j*3+1], this.cols[j*3+2]);
			}
		}

		for (int iRow = 0; iRow < 9; iRow++) {
			for (int iCol = 0; iCol < 9; iCol++) {
				this.cells[iRow][iCol].makeReferences(this.rows[iRow], this.cols[iCol], this.sqrs[iRow/3][iCol/3]);
				this.session.insert(this.cells[iRow][iCol]);
			}
			this.session.insert(this.rows[iRow]);
			this.session.insert(this.cols[iRow]);
			this.session.insert(this.sqrs[iRow/3][iRow%3]);
		}
	}

	public void setCellValues(final Integer[][] cellValues) {
		if (this.session != null) {
			this.session.dispose();
			this.steppingFactHandle = null;
		}

		this.session = this.kc.newKieSession("SudokuKS");
		this.session.setGlobal("explain", this.explain);

		final Setting s000 = new Setting(0, 0, 0);
		final FactHandle fh000 = this.session.insert(s000);
		this.create();

		int initial = 0;
		for (int iRow = 0; iRow < 9; iRow++) {
			for (int iCol = 0; iCol < 9; iCol++) {
				final Integer value = cellValues[iRow][iCol];
				if (value != null) {
					this.session.insert(new Setting(iRow, iCol, value));
					initial++;
				}
			}
		}
		this.counter = new Counter(initial);
		this.session.insert(this.counter);
		this.session.delete(fh000);
		this.session.fireAllRules();
	}

	@Override
	public String toString() {
		final StringBuilder sb = new StringBuilder();

		sb.append("Sudoku:").append('\n');
		for (int iRow = 0; iRow < 9;  iRow++) {
			sb.append("  ").append(this.rows[iRow].toString()).append('\n');
		}

		return sb.toString();
	}

	public void validate(){
		this.session.getAgenda().getAgendaGroup( "validate" ).setFocus();
		this.session.fireUntilHalt();
	}
}
